package jogl;

import java.awt.*;
import java.awt.event.*; 

import net.java.games.jogl.*;
import net.java.games.jogl.util.GLUT;

import javax.vecmath.*;
// import javax.swing.*;

public class ArcBall extends Frame implements GLEventListener, MouseListener, MouseMotionListener {
    
	Vector2d translation = new Vector2d();
	
	public GLCanvas canvas;
	public static final Dimension PREFERRED_FRAME_SIZE =
        new Dimension (800,800);
    
    public ArcBall() {
    	// init Frame
        super ("JOGL!");
        // get a GLCanvas
        GLCapabilities capabilities = new GLCapabilities();
        canvas =
            GLDrawableFactory.getFactory().createGLCanvas(capabilities);
        // add a GLEventListener, which will get called when the
        // canvas is resized or needs a repaint
        canvas.addGLEventListener(this);
        canvas.addMouseListener(this);
        canvas.addMouseMotionListener(this);
        // now add the canvas to the Frame.  Note we use BorderLayout.CENTER
        // to make the canvas stretch to fill the container (ie, the frame)
        add (canvas, BorderLayout.CENTER);
        this.addWindowListener(new java.awt.event.WindowAdapter() {

            public void windowClosing(WindowEvent e) {
            	System.out.println("closing");
            	System.exit(0);
            }
          });     
    }
    
    /** We'd like to be 600x400, please.
     */
    public Dimension getPreferredSize () {
        return PREFERRED_FRAME_SIZE;
    }
    
    /** main just creates and shows a JOGL2DBasics Frame
     */
    public static void main (String[] args) {        
    	Frame f = new ArcBall();
        f.pack();
        f.setVisible(true);
    }
    
    
    /*
     * METHODS DEFINED BY GLEventListener
     */
    
    /** Called by drawable to initiate drawing 
     */
    public void display (GLDrawable drawable) {
        GL gl = drawable.getGL();
        GLU glu = drawable.getGLU();
        GLUT glut = new GLUT();
        setCam(gl);
        
        gl.glMatrixMode( GL.GL_MODELVIEW );
        gl.glLoadIdentity();
        /*
        gl.glLoadIdentity();
        vecMath2GL(transform, trans);
        //System.out.println(transform);
        gl.glLoadMatrixd(trans);
        */
        gl.glLoadMatrixd(trans);
        gl.glTranslated(translation.x, translation.y, 0);
        draw(gl, glu, glut);
    }
    
    public void draw(GL gl, GLU glu, GLUT glut)
    {    	
    	gl.glClear (GL.GL_COLOR_BUFFER_BIT);
        gl.glClear (GL.GL_DEPTH_BUFFER_BIT);
        gl.glColor3f( 1.0f, 0.0f, 0.0f ); 
        gl.glLoadIdentity();
        gl.glLoadMatrixd(trans);
        
        glut.glutSolidCube(gl, 2);    	    
    }
    
    public void vecMath2GL(Matrix4d M, double[]m){
    	m[0] = M.m00;
    	m[1] = M.m10;
    	m[2] = M.m20;
    	m[3] = M.m30;

    	m[4] = M.m01;
    	m[5] = M.m11;
    	m[6] = M.m21;
    	m[7] = M.m31;

    	m[8] = M.m02;
    	m[9] = M.m12;
    	m[10] = M.m22;
    	m[11] = M.m32;    	

    	m[12] = M.m03;
    	m[13] = M.m13;
    	m[14] = M.m23;
    	m[15] = M.m33;
    }
    
    /** Called by drawable to indicate mode or device has changed
     */
    public void displayChanged (GLDrawable drawable,
                                boolean modeChanged,
                                boolean deviceChanged) {
        System.out.println ("displayChanged()");
    }
    
    public Matrix4d transform;
    
    Matrix4d lastTransform;
    Matrix4d lastTransformInv;
    public double trans[];        
    
    /** Called after OpenGL is init'ed
     */
    public void init (GLDrawable drawable) {
        //System.out.println ("init()");
        transform = new Matrix4d();
        transform.setIdentity();
        lastTransform = new Matrix4d();
        lastTransform.setIdentity();
        lastTransformInv = new Matrix4d();
        lastTransformInv.setIdentity();
        trans = new double[]{1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1};
        GL gl = drawable.getGL(); 
        // set erase color
        gl.glClearColor( 1.0f, 1.0f, 1.0f, 1.0f ); //white 
        // set drawing color and point size
        gl.glPointSize(4.0f); //a 'dot' is 4 by 4 pixels 
        gl.glEnable(GL.GL_LIGHTING);
        gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, new float[]{0, 0, 1, 0});
        gl.glEnable(GL.GL_LIGHT0);
        /*
        gl.glEnable(GL.GL_LIGHT1);
        gl.glLightfv(GL.GL_LIGHT1, GL.GL_POSITION, new float[]{1, 1, 1, 0});
        */
        
        gl.glEnable(GL.GL_DEPTH_TEST);
        gl.glColorMaterial(GL.GL_FRONT, GL.GL_DIFFUSE);
        gl.glEnable(GL.GL_COLOR_MATERIAL);
		//gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE);        
		gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_FILL);
    }
    
    public double camWidth = .5;
    public double camHeight = .5;
    
    /** Called to indicate the drawing surface has been moved and/or resized
     */
    public void reshape (GLDrawable drawable,
                         int x,
                         int y,
                         int width,
                         int height) {
        //System.out.println ("reshape()");
        GL gl = drawable.getGL(); 
        GLU glu = drawable.getGLU(); 
        gl.glViewport( 0, 0, width, height ); 
        setCam(gl);
        /*
        gl.glMatrixMode( GL.GL_PROJECTION );  
        gl.glLoadIdentity(); 
        //glu.gluOrtho2D( 0.0, 450.0, 0.0, 375.0);
        gl.glOrtho(-camWidth/2, camWidth/2, -camHeight/2, camHeight/2, -30, 30);
        */
    }
    
    private void setCam(GL gl){
        gl.glMatrixMode( GL.GL_PROJECTION );  
        gl.glLoadIdentity(); 
        //glu.gluOrtho2D( 0.0, 450.0, 0.0, 375.0);
        gl.glOrtho(-camWidth/2, camWidth/2, -camHeight/2, camHeight/2, -30, 30);    
        gl.glMatrixMode(GL.GL_MODELVIEW);
    }
    

	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	public void mouseClicked(MouseEvent arg0) {
		// TODO Auto-generated method stub
		
	}

	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 */
	public void mouseEntered(MouseEvent arg0) {
		// TODO Auto-generated method stub
		
	}

	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 */
	public void mouseExited(MouseEvent arg0) {
		// TODO Auto-generated method stub
		
	}

	public void screen2sphere(int x, int y, Vector3d sphere){
		Dimension canvasSize = canvas.getSize();
		double nx = camWidth * x / canvasSize.width - camWidth/2;
		double ny = -camHeight * y / canvasSize.height + camHeight/2;
		
		double sphereRadius = camWidth / 2;
		
		if(nx*nx + ny*ny <= sphereRadius*sphereRadius){
			sphere.x = nx;
			sphere.y = ny;
			sphere.z = Math.sqrt(sphereRadius*sphereRadius - nx*nx - ny*ny);
		}
		else{
			sphere.x = nx;
			sphere.y = ny;
			sphere.z = 0;
			sphere.normalize();
		}
	}
	
	int lastMouseX, lastMouseY;
	boolean rotate;
	boolean translate;
	boolean zoom = false;
	
	Vector3d rotateFrom = new Vector3d();
	Vector3d rotateTo = new Vector3d();
	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	public void mousePressed(MouseEvent arg0) {
		// TODO Auto-generated method stub
		lastMouseX = arg0.getX();
		lastMouseY = arg0.getY();
		//System.out.println(arg0.getButton());
		if(arg0.getButton() == MouseEvent.BUTTON1){
			rotate = true;
			translate = false;
			zoom = false;
			screen2sphere(lastMouseX, lastMouseY, rotateFrom);
		}
		else if(arg0.getButton() == MouseEvent.BUTTON3){
			rotate = false;
			translate = false;
			zoom = true;
		}
		else{
			rotate = false;
			translate = true;
			zoom = false;			
		}
		/*
		else if(arg0.getButton() == MouseEvent.BUTTON3){
			rotate = false;
			translate = true;
		}*/
		//arg0.consume();
	}

	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	public void mouseReleased(MouseEvent arg0) {
		// TODO Auto-generated method stub
		saveTransform();
		//arg0.consume();
	}

	public void saveTransform(){
		lastTransform.transform(sceneCenter);
		lastTransform.set(transform);
	}
	
	/* (non-Javadoc)
	 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
	 */
	public void mouseDragged(MouseEvent arg0) {
		// TODO Auto-generated method stub
		int dx = arg0.getX() - lastMouseX;
		int dy = -(arg0.getY() - lastMouseY);
		if(rotate){
			screen2sphere(arg0.getX(), arg0.getY(), rotateTo);
			rotate(rotateFrom, rotateTo);
			//rotate(dx, dy);
		}
		else if(translate){
			translate(dx, dy);
		}
		else if(zoom){
			camWidth *= Math.pow(1.1, ((float)dy)/10);
			camHeight = camWidth;
			lastMouseX = arg0.getX();
			lastMouseY = arg0.getY();
		}
		else if(translate){
			translation.x += ((float)dx) /100; 
			translation.y += ((float)dy) /100; 
			lastMouseX = arg0.getX();
			lastMouseY = arg0.getY();		
		}
		//lastMouseX = arg0.getX();
		//lastMouseY = arg0.getY();
        vecMath2GL(transform, trans);
        canvas.repaint();
        //arg0.consume();
	}
	
	Vector3d tempVec = new Vector3d();
	public void rotate(Vector3d from, Vector3d to){
		tempVec.cross(from, to);
		tempVec.normalize();
		tempMatrix.setIdentity();
		tempMatrix.setRotation(new AxisAngle4d(tempVec, Math.abs(from.angle(to))));
		tempMatrixInv.setIdentity();
		tempMatrixInv.setRotation(new AxisAngle4d(tempVec, -Math.abs(from.angle(to))));
		transform.set(tempMatrix);
		transform.mul(lastTransform);
		
	}
	
	Vector3d sceneCenter = new Vector3d(0, 0, 0);
	Matrix4d tempMatrix = new Matrix4d();
	Matrix4d tempMatrixInv = new Matrix4d();
	public void rotate(double dx, double dy){
		tempMatrix.setIdentity();
		Vector3d tempVec = new Vector3d(dx, dy, 0);
		Vector3d tempVec3 = new Vector3d(0, 0, 1);
		tempVec.cross(tempVec3, tempVec);
		tempVec.normalize();
		tempMatrix.setRotation(new AxisAngle4d(tempVec, .01*Math.sqrt(dx*dx + dy*dy)));
		transform.set(tempMatrix);
		transform.mul(lastTransform);
		
	}

	public void translate(double dx, double dy){
		tempMatrix.setIdentity();
		Dimension canvasSize = canvas.getSize();
		tempMatrix.setTranslation(new Vector3d(camWidth * ((double)dx) / canvasSize.width, camHeight * ((double)dy) / canvasSize.height, 0)); 
		//System.out.println(tempMatrix);

		transform.set(tempMatrix);
		transform.mul(lastTransform);
		/*
		tempMatrix.mul(transform);
		Matrix4d temp = transform;
		transform = tempMatrix;
		tempMatrix = temp;
		Sytem.out.println(transform);*/
	}
	
	/* (non-Javadoc)
	 * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
	 */
	public void mouseMoved(MouseEvent arg0) {
		// TODO Auto-generated method stub
		
	}        
}
