/*
 * Created on Jul 18, 2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jogl;

import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.io.*;
import java.util.*;

import javax.imageio.ImageIO;
import javax.imageio.ImageWriter;
import javax.vecmath.Vector2d;

import Jama.EigenvalueDecomposition;
import Jama.Matrix;

import net.java.games.jogl.GL;

/**
 * @author golovins
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class Image {
	WrinkleVector wrinkleVector;
	
	
	final static int ABSOLUTE = 0;
	final static int MIN_MAX = 1;
	final static int AVG_2STD = 2;
	final static int AVG_STD = 3;
	
	int drawScale = MIN_MAX;
	
	Statistics statistics[][];
	int statisticsSize = -1;
	
	SimoncelliPyramid pyramid;
	
	int width, height;
	Image reflection;
	float minValue = Float.POSITIVE_INFINITY, maxValue = Float.NEGATIVE_INFINITY;
	
	float[][] data;
	
	float[] drawnImage;
	
	Image[] lowPasses;
	Image[] details;
	
	float drawnMin, drawnMax;
	
	String fileName;
	
	public Image(int width, int height, String file){
		this(width, height);
		try{
			BufferedImage bufferedImage = ImageIO.read(new File(file));
			System.out.println("Loaded image: " + bufferedImage.getHeight() + "x" + bufferedImage.getWidth());
			ColorModel colorModel = ColorModel.getRGBdefault();
			int bwidth = bufferedImage.getWidth();
			int bheight = bufferedImage.getHeight();
			
			for(int i = 0; i < width; i++)
				for(int j = 0; j < height; j++)
					data[i][j] = ((float)colorModel.getRed(bufferedImage.getRGB(i %bwidth, j % bheight))) / 255;
		}
		catch(Exception e){
			e.printStackTrace();
		}

	}
	
	public Image(int width, int height, String files[]){
		this(width, height);
		int tiles = (int)Math.round(Math.sqrt(files.length)); 
		if(tiles * tiles != files.length)
			throw new RuntimeException("Not a square");
		Image images[] = new Image[files.length];
		for(int i = 0; i < images.length; i++)
			images[i] = new Image(width / tiles, height / tiles, files[i]);
		int smallImageSize = width / tiles;
		
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++){
				int x = i / smallImageSize;
				int y = j / smallImageSize;
				data[i][j] = images[x + y * tiles].data[i % smallImageSize][j % smallImageSize];
			}
	}
		
	public Image(int width, int height){
		this.width = width;
		this.height = height;
		data = new float[width][height];
		lowPasses = new Image[1];
		lowPasses[0] = this;
		details = new Image[0];
	}
	
	public Image(double[] imageArray, int offset){
		this((int)Math.round(imageArray[0 + offset]), (int)Math.round(imageArray[1 + offset]));
		for(int x = 0; x < width; x++){
			for(int y = 0; y < height; y++){
				data[x][y] = (float)imageArray[y * width + x + offset + 2];
			}					
		}
	}
	
	public Image(Image image){
		this(image.width, image.height);
		fileName = 2 + image.fileName;
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				data[i][j] = image.data[i][j];
		//createDrawnImage(true);
	}
	
	public Image(String file){
		try{
			fileName = file;
			int fileLength = (int)(new File(file)).length();
			byte fileBytes[] = new byte[fileLength];
			FileInputStream fStream = new FileInputStream(file);
			fStream.read(fileBytes);
			fStream.close();
			
			DataInputStream dstream = new DataInputStream(new ByteArrayInputStream(fileBytes));
			width = dstream.readInt();
			height = dstream.readInt();
			data = new float[width][];
			for(int x = 0; x < width; x++){
				data[x] = new float[height];
			}
			
			for(int y = 0; y < height; y++){
				for(int x = 0; x < width; x++){
					data[x][y] = dstream.readFloat();
					if(data[x][y] > maxValue) maxValue = data[x][y];
					if(data[x][y] < minValue) minValue = data[x][y];
				}					
			}
			createDrawnImage(true);
			System.out.println("Loaded " + file + "; " + width + "x" + height + ", min/max: " + minValue + "/" + maxValue);
		}
		catch(IOException e){
			e.printStackTrace();
		}
		
	}
	
	public void saveAs(String file){
		saveAs(file, false);
	}
	
	private static boolean closeToOne(float f){
		return Math.abs(f - 1) < .001;
	}
	
	public static void setBufferedImagePixel(BufferedImage image, int x, int y, float r, float g, float b){
		ColorModel cmodel = ColorModel.getRGBdefault();
		float[] components = new float[]{r, g, b, 1};
		
		
		if(closeToOne(r) && closeToOne(g) && closeToOne(b)){
			components[3] = 0;
		}
		
		
		int color = cmodel.getDataElement(components, 0);
		image.setRGB(x, image.getHeight() - y - 1, color);

	}
	
	public void saveAs(String file, boolean normalize){
		BufferedImage bimage = new BufferedImage(width, height,
				BufferedImage.TYPE_INT_RGB);
		float maxValue = 1;
		float minValue = 0;
		if(normalize){
			maxValue = Float.NEGATIVE_INFINITY;
			minValue = Float.POSITIVE_INFINITY;
			for(int i = 0; i < width; i++){
				for(int j = 0; j < height; j++){
					if(data[i][j] > maxValue)
						maxValue = data[i][j];
					if(data[i][j] < minValue)
						minValue = data[i][j];
				}
			}
			for(int i = 0; i < width; i++)
				for(int j = 0; j < height; j++)
					data[i][j] = (data[i][j] - minValue) / (maxValue - minValue);
		}
		
		for(int i = 0; i < width; i++){
			for(int j = 0; j < height; j++){
				setBufferedImagePixel(bimage, i, j, 
						data[i][j], data[i][j], data[i][j]);
			}
		}
		
		try{
			
			ImageIO.write(bimage, "bmp", new File(file));
			System.out.println("Image saved to: " + file + "; normalized: " + normalize);
		}
		catch(Exception e){
			e.printStackTrace();
		}
	}
	
	public void createDrawnImage(float mins[], float maxes[]){
		createDrawnImage(mins[mins.length - 1], maxes[maxes.length - 1]);
		if(pyramid != null){
			for(int i = 0; i < pyramid.bands.length; i++)
				pyramid.bands[i].createDrawnImage(mins[i], maxes[i]);
		}
	}
	
	public void createDrawnImage(float min, float max){
		drawnMin = min;
		drawnMax = max;
		
		drawnImage = new float[width*height];
		for(int x = 0; x < width; x++){
			for(int y = 0; y < height; y++){
				drawnImage[y * width + x] = data[x][y];
			}					
		}
		for(int i = 0; i < drawnImage.length; i++)
			drawnImage[i] = (drawnImage[i] - min) / (max - min);		
	}
	
	public void createDrawnImage(){
		if(drawScale == ABSOLUTE)
			createDrawnImage(false);
		else
			createDrawnImage(true);
	}
	
	public void createDrawnImage(boolean normalize){
		drawnImage = new float[width*height];
		for(int x = 0; x < width; x++){
			for(int y = 0; y < height; y++){
				drawnImage[y * width + x] = data[x][y];
			}					
		}
		if(normalize){
			float max = Float.NEGATIVE_INFINITY, min = Float.POSITIVE_INFINITY;
			for(int i = 0; i < drawnImage.length; i++){
				if(drawnImage[i] > max) max = drawnImage[i];
				if(drawnImage[i] < min) min = drawnImage[i];
			}
			float avg = (max + min)/2;
			float std = (float)Math.sqrt(var());
			float enhancement = 0f; 
			
			/*
			max = max * (1 - enhancement) + avg * enhancement;
			min = min * (1 - enhancement) + avg * enhancement;
			*/
			if(drawScale == AVG_STD){
				avg = avg();
				max = avg + std;
				min = avg - std;
			}
			else if(drawScale == AVG_2STD){
				avg = avg();
				max = avg + 2f * std;
				min = avg - 2f * std;				
			}
				
			//System.out.println("   std: " + std);
			//System.out.println("Normalizing displayed image to min: " + min + ", max: " + max);				
			for(int i = 0; i < drawnImage.length; i++)
				drawnImage[i] = (drawnImage[i] - min) / (max - min);
		}
	}
	
	float getLPImpulse(int m, int n){
		return 1; 
	}
	
	Image getLowPass(){
		if(width % 2 != 0 || height % 2 != 0)
			throw new RuntimeException("Width or height not even");
		
		Image result = new Image(width/2, height/2);
		for(int i = 0; i < result.width; i++){
			for(int j = 0; j < result.height; j++){
				//System.out.println(i + " " + j);
				result.data[i][j] = 0;
				float normalization = 0;
				for(int m = -2; m <=2; m++){
					int x = 2*i + m;
					if(x < 0) continue;
					if(x >= width) break;
					for(int n = -2; n <= 2; n++){
						int y = 2*j + n;
						if(y < 0) continue;
						if(y >= height) break;
						float weight = getLPImpulse(m, n);
						//System.out.println(m + " " + n + ": " + weight);
						normalization += weight;
						result.data[i][j] += weight * data[x][y];
					}
				}
				result.data[i][j] /= normalization;
				/*
				for(int x = 2*i; x < 2*i + 2; x++){
					for(int y = 2*j; y < 2*j + 2; y++){
						result.data[i][j] += .25 * data[x][y];
					}
				}
				*/
			}
		}
		result.createDrawnImage(true);
		return result;
	}
	
	Image getDetails(Image lowPass){
		if(lowPass.width != width / 2 || lowPass.height != height / 2)
			throw new RuntimeException("Width or height of lowpass incorrect");
		Image result = new Image(width, height);
		for(int i = 0; i < width; i++){
			for(int j = 0; j < height; j++){
				result.data[i][j] = data[i][j] - lowPass.data[i/2][j/2];
				//result.data[i][j] = 1;
			}
		}
		result.createDrawnImage(true);
		return result;
	}
	
	void createPyramid(){
		int levels = 6;
		lowPasses = new Image[levels];
		details = new Image[levels - 1]; 
		lowPasses[0] = this;
		for(int i = 1; i < lowPasses.length; i++){
			lowPasses[i] = lowPasses[i-1].getLowPass();
			details[i - 1] = lowPasses[i - 1].getDetails(lowPasses[i]);
		}
	}
	
	void createSteerablePyramid(int orientations, int levels){
		pyramid = new SimoncelliPyramid(orientations);
		pyramid.expand(this, levels, new CircularEdgeHandler());
		pyramid.createHistograms(ImageLoader.numBins, ImageLoader.numStatisticTiles);
		/*
		createHistograms(true, 256);
		for(int i = 0; i < pyramid.bands.length; i++)
			pyramid.bands[i].createHistograms(true, 256);
		*/ 
		//pyramid.colla
	}
	
	void drawImage(GL gl){
		gl.glDrawPixels(width, height, GL.GL_LUMINANCE, GL.GL_FLOAT, drawnImage);
	}

	void drawImage(GL gl, int x, int y){						
		drawImage(gl, x, y, false);
	}
	
	void drawImage(GL gl, int x, int y, boolean drawCDF){						
		gl.glRasterPos2i(x, y);
		if(drawnImage == null)
			createDrawnImage(true);
		gl.glDrawPixels(width, height, GL.GL_LUMINANCE, GL.GL_FLOAT, drawnImage);
		
		if(statistics != null && drawCDF){
			gl.glDisable(GL.GL_DEPTH_TEST);
			for(int i = 0; i < statisticsSize; i++)
				for(int j = 0; j < statisticsSize; j++){
					int pixelX = x + i * width / statisticsSize;
					int pixelY = y + j * height / statisticsSize /*+ height*/;
					statistics[i][j].CDF.draw(gl, pixelX, pixelY, width / statisticsSize, drawnMin, drawnMax, this);
				}
		}
		
	}
	
	void draw(GL gl, int x, int y, boolean drawCDF){
		if(pyramid != null)
			pyramid.draw(gl, x, y, drawCDF);
		else
			drawImage(gl, x, y, drawCDF);
	}
	
	void drawPyramid(GL gl, int x, int y){
		int currentX = x, currentY = y;
		for(int i = 0; i < lowPasses.length; i++){
			lowPasses[i].drawImage(gl, currentX, currentY);						
			
			if(i < details.length)
				details[i].drawImage(gl, currentX, currentY + height);
			
			currentX += lowPasses[i].width + 2;
		}
	}
	
	public float data(float x, float y){
		int xInt = (int)(x);
		int yInt = (int)(y);	        		
		int xRight = xInt + 1;
		int yUp = yInt + 1;
		
		if(xInt < 0){
			//System.out.println("warning: " + x + ", " + y);
			xInt = 0;
		}
		if(yInt < 0){
			//System.out.println("warning: " + x + ", " + y);
			yInt = 0;
		}
		if(xRight >= width){
			//System.out.println("warning: " + x + ", " + y);
			xRight = width -1;
		}
		if(yUp >= height){
			//System.out.println("warning: " + x + ", " + y);
			yUp = height -1;
		}

		
		float xRemainder = x - xInt;
		float yRemainder = y  - yInt;
		
		float downLeft = data[xInt][yInt];
		float downRight = data[xRight][yInt];
		float upLeft = data[xInt][yUp];
		float upRight = data[xRight][yUp];

		float left = downLeft * (1 - yRemainder) + upLeft * yRemainder;
		float right = downRight * (1 - yRemainder) + upRight * yRemainder;
	
		return left * (1 - xRemainder) + right * xRemainder;		
	}
	
	public void save(String imageFile){
		try{
			FileOutputStream fstream = new FileOutputStream(imageFile, false);
			BufferedOutputStream bstream = new BufferedOutputStream(fstream, width * height * 4);
			DataOutputStream dstream = new DataOutputStream(bstream);				
			dstream.writeInt(width);
			dstream.writeInt(height);
			float[] prevDrawnImage = drawnImage;
			createDrawnImage(false);
			for(int i = 0; i < drawnImage.length; i++){
				dstream.writeFloat(drawnImage[i]);
			}
							
			dstream.close();
			System.out.println("Saved " + width + "x" + height + " image to: " + imageFile);
			drawnImage = prevDrawnImage;
		}
		catch(IOException e){
			e.printStackTrace();
		}
	}
	
	public void interpolate(Image image, float distance){
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				data[i][j] = 2 * (data[i][j] * (1 - distance) + image.data[i][j] * distance);
	}
	
	public Image reflect(){
		//System.out.println("  reflecting");
		Image image = new Image(width, height);
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				image.data[i][j] = data[width - 1 - i][height - 1 - j];
		return image;
	}
	
	static public void convolve(Image I, Image filter, Image R, boolean forward,  int rStep, int cStep){
		convolve(I, filter, R, forward,  rStep, cStep, null);
	}
	
	static public void convolve(Image I, Image filter, Image R, boolean forward,  int rStep, int cStep, EdgeHandler EH){
		if(R == null || R.width != I.width / rStep || R.height != I.height / rStep)
			throw new RuntimeException("Result image bad");
		if(R == I)
			throw new RuntimeException("Target and source of convolution same");
		
		if(EH == null)
			EH = new ZeroEdgeHandler();
		
		Image F = filter;
		if(!forward){
			if(filter.reflection == null)
				filter.reflection = filter.reflect();
			F = filter.reflection;
		}
		
		//System.out.println(F);
		
		int rCenter = F.width / 2;
		int cCenter = F.height / 2;

		float imageData;
		int xConv, yConv;
		for(int x = 0, xR = 0; x < I.width; x+= rStep, xR++){
			for(int y = 0, yR = 0; y < I.height; y+= cStep, yR++){
				//System.out.println();
				//System.out.println(x + ", " + y);
				R.data[xR][yR] = 0;
				for(int i = 0; i < F.width; i++){
					xConv = x - rCenter + i;
					//if(xConv < 0 || xConv >= I.width) continue;
					for(int j = 0; j < F.height; j++){						
						yConv = y - rCenter + j; 
						//if(yConv < 0 || yConv >= I.height) continue;
						//System.out.println(xConv + ", " + yConv + "; " + i + ", " + j + ": " + F.data[i][j]);
						
						if(xConv < 0 || xConv >= I.width || yConv < 0 || yConv >= I.height)
							imageData = EH.get(xConv, yConv, I);
						else{
							imageData = I.data[xConv][yConv];
						}
						
						R.data[xR][yR] += imageData * F.data[i][j];
					}
				}
			}
		}
	}
	
	static void upSample(Image I, Image R)
	{
		if(R == null || R.width != I.width * 2 || R.height != I.height * 2)
			throw new RuntimeException("Result of bad dimension");
		for(int i = 0; i < R.width; i++)
			for(int j = 0; j < R.height; j++)
				R.data[i][j] =0;
		
		for(int i = 0; i < I.width; i++)
			for(int j = 0; j < I.height; j++)
				R.data[i*2][j*2] = I.data[i][j];
	}
	
	static void add(Image A, Image B, Image result){
		if(A.width != B.width || A.width != result.width || 
				A.height != B.height || A.height != result.height)
			throw new RuntimeException("Dimensions not equal");
		for(int i = 0; i < A.width; i++)
			for(int j = 0; j < A.height; j++)
				result.data[i][j] = A.data[i][j] + B.data[i][j];
	}

	
	static Image incline(int size, float low, float high){
		//float low = .4f, high = 1f;
		Image image = new Image(size, size);
		for(int i = 0; i < size; i++)
			for(int j = 0; j < size; j++)
				image.data[i][j] = ((float)i) / size * (high - low) + low;
		image.createDrawnImage();
		image.fileName = "incline.img";
		return image;
	}
	
	static Image randomImage(int width, int height, float min, float max){
		Random rand = new Random();
		Image image = new Image(width, height);
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				image.data[i][j] = rand.nextFloat() * (max - min) + min;
		return image;
	}

	static Image gaussianImage(int width, int height, float avg, float std){
		Random rand = new Random();
		Image image = new Image(width, height);
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				image.data[i][j] = (float)rand.nextGaussian() * std + avg;
		return image;
	}
	
	public String toString(){
		String output = "";
		for(int j = 0; j < height; j++){
			output += "\n";
			for(int i = 0; i < width; i++){
				output += data[i][j] + "  ";
			}
		}
		return output;
	}
	
	public float sum(int startX, int startY, int endX, int endY){
		float sum = 0;
		for(int i = startX; i < endX; i++)
			for(int j = startY; j < endY; j++)
				sum += data[i][j];
		return sum;
	}
	
	public float sumSquared(int startX, int startY, int endX, int endY){
		float sumSquared = 0;
		for(int i = startX; i < endX; i++)
			for(int j = startY; j < endY; j++)
				sumSquared += data[i][j] * data[i][j];
		return sumSquared;
	}
	
	public float avg(int startX, int startY, int endX, int endY){
		return sum(startX, startY, endX, endY) / (endX - startX) / (endY - startY);	
	}
	
	public float var(int startX, int startY, int endX, int endY){
		float avg = avg(startX, startY, endX, endY);
		return sumSquared(startX, startY, endX, endY) / (endX - startX) / (endY - startY) - avg * avg;	
	}
	
	public float avg(){
		return avg(0, 0, width, height);
	}
	
	public float var(){
		return var(0, 0, width, height);
	}
	
	public float min(int startX, int startY, int endX, int endY){
		float min = Float.POSITIVE_INFINITY;
		for(int i = startX; i < endX; i++)
			for(int j = startY; j < endY; j++)
				if(data[i][j] < min)
					min = data[i][j];
		return min;
	}

	public float max(int startX, int startY, int endX, int endY){
		float max = Float.NEGATIVE_INFINITY;
		for(int i = startX; i < endX; i++)
			for(int j = startY; j < endY; j++)
				if(data[i][j] > max)
					max = data[i][j];
		return max;
	}		

	public float min(){
		return min(0, 0, width, height);
	}

	public float max(){
		return max(0, 0, width, height);
	}		
		
	public static boolean isBadFloat(float f){
		return (f == Float.NaN || f == Float.NEGATIVE_INFINITY ||
				f == Float.POSITIVE_INFINITY);
	}
	
	float getStatisticsPrediction(int i, int j, int x, int y, Object[][] stats /* Image sourceImage*/){
		if(stats instanceof InverseCDF[][]){			
			InverseCDF[][] ICDF = (InverseCDF[][])stats; 
			int binIndex = statistics[x][y].PDF.indexOf(data[i][j]);
			float cumulativeProb = statistics[x][y].CDF.counts[binIndex];
			return /*sourceImage.statistics[x][y].*/ICDF[x][y].getValue(cumulativeProb);
		}
		else if(stats instanceof ShortStats[][]){
			//return data[i][j];
			ShortStats[][] shortStats = (ShortStats[][]) stats;
			float sourceMean = statistics[x][y].shortStats.mean;
			float targetMean = shortStats[x][y].mean;
			
			float sourceSTD = statistics[x][y].shortStats.std;
			float targetSTD = shortStats[x][y].std;
			
			
			if(isBadFloat(sourceMean) || isBadFloat(targetMean) ||
					isBadFloat(sourceSTD) || isBadFloat(targetSTD) ||
					sourceSTD <= 0 || targetSTD <= 0)
				throw new RuntimeException();
			
			
			//System.out.print("[" + targetMean + "; " + sourceMean + "]" + " [");
			//System.out.print(targetSTD + "; " + sourceSTD + "] \n");
			
			//targetSTD = targetSTD / 2;
			return (data[i][j] - sourceMean) * targetSTD / sourceSTD + sourceMean /*targetMean*/;
		}
		else
			throw new RuntimeException();
	}
	
	public Statistics getStatistics(int i, int j){
		int x = i * statisticsSize / width;
		int y = j * statisticsSize / height;
		return statistics[x][y];
	}
		
	public void matchHistogram(/*Image image*/Object[][] stats, int numStatistics){
		/*
		if(width != height)
			throw new RuntimeException();
		*/
		if(stats.length != numStatistics || stats[0].length != numStatistics)
			throw new RuntimeException();
		
		int bins = ImageLoader.numBins;
		createHistograms(bins, numStatistics);
		
		//image.createHistograms(bins, numStatistics);

		int cellSizeX = width / statisticsSize;
		int cellSizeY = height / statisticsSize;
		for(int i = 0; i < width; i++){
			for(int j = 0; j < height; j++){
				
				int xLow = (i - cellSizeX / 2) / cellSizeX;
				int yLow = (j - cellSizeY / 2) / cellSizeY;
				int xHigh = xLow + 1;
				int yHigh = yLow + 1;
				if(xLow < 0) xLow = 0;
				if(yLow < 0) yLow = 0;
				if(xHigh >= statisticsSize) xHigh = statisticsSize - 1;
				if(yHigh >= statisticsSize) yHigh = statisticsSize - 1;

				float xRemainder = ((((float)i) - cellSizeX / 2) / cellSizeX - xLow);
				float yRemainder = ((((float)j) - cellSizeY / 2) / cellSizeY - yLow);
				
				float lowerLeft = getStatisticsPrediction(i, j, xLow, yLow, stats/*image*/);
				float lowerRight = getStatisticsPrediction(i, j, xHigh, yLow, stats/*image*/);
				float upperLeft = getStatisticsPrediction(i, j, xLow, yHigh, stats/*image*/);
				float upperRight = getStatisticsPrediction(i, j, xHigh, yHigh, stats/*image*/);
				
				float left = lowerLeft * (1 - yRemainder) + upperLeft * yRemainder;
				float right = lowerRight * (1 - yRemainder) + upperRight * yRemainder;
				
				data[i][j] = left * (1 - xRemainder) + right * xRemainder;
				
				//System.out.println(i + " " + j + ": " + xLow + ", " + yLow + "; " + xRemainder + ", " + yRemainder);
				
				/*
				int x = i * statisticsSize / width;
				int y = j * statisticsSize / height;
				data[i][j] = getStatisticsPrediction(i, j, x, y, image);
				*/
				
			}
		}
		/*
		for( int i = 0; i < Target.Size(); i++ )
		{
			Idx = (int)((Target(i) - Min)/BinSize);
	        if (Idx < 0) Idx = 0;
	        if (Idx > Bins-1) Idx = Bins-1;
			Val = TCDF[0][Idx];
			Idx = (int)(Size * Val);
			Target(i) = SICDF[1][Idx];
		}
		*/
	}
	
	public void createHistograms(int bins, int numStatistics){
		statistics = new Statistics[numStatistics][numStatistics];
		statisticsSize = numStatistics;
		for(int i = 0; i < statisticsSize; i++){
			for(int j = 0; j < statisticsSize; j++){
				int startX = i * width / statisticsSize;
				int endX = (i + 1) * width / statisticsSize;
				int startY = j * height / statisticsSize;
				int endY = (j + 1) * height / statisticsSize;
				statistics[i][j] = new Statistics(this, startX, startY, endX, endY);
			}
		}
	}
	
	public void copyFrom(Image image){
		if(width != image.width || height != image.height){
			throw new RuntimeException("Incorrect size");
		}
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				data[i][j] = image.data[i][j];
	}
	
	void addNoise(){
		Random random = new Random();
		float std = (float)Math.sqrt(var());
		
		/*
		for(int i = 0; i < width; i++)
			for(int j = 0; j < width; j++)
				data[i][j] += random.nextGaussian() * std / 20; //10; //!!!!!!!!!!
		*/
		
		addPerlinNoise(std);
		
		createHistograms(ImageLoader.numBins, ImageLoader.numStatisticTiles);
		/*
		System.out.println(statisticsSize + ", " + statBoxSize);
		for(int i = 0; i < statistics.length; i++)
			for(int j = 0; j < statistics[i].length; j++)
				System.out.println(statistics[i][j].std);
		*/
		
		/*
		int statBoxSize = width / statisticsSize;
		Random random = new Random();
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++){
				int xLow = (i - statBoxSize / 2) / statBoxSize;
				int yLow = (j - statBoxSize / 2) / statBoxSize;
				int xHigh = xLow + 1;
				int yHigh = yLow + 1;
				if(xLow < 0) xLow = 0;
				if(yLow < 0) yLow = 0;
				if(xHigh >= statisticsSize) xHigh = statisticsSize - 1;
				if(yHigh >= statisticsSize) yHigh = statisticsSize - 1;

				float xRemainder = ((((float)i) - statBoxSize / 2) / statBoxSize - xLow);
				float yRemainder = ((((float)j) - statBoxSize / 2) / statBoxSize - yLow);
				
				float lowerLeft = statistics[xLow][yLow].shortStats.std;
				float lowerRight = statistics[xHigh][yLow].shortStats.std;
				float upperLeft = statistics[xLow][yHigh].shortStats.std;
				float upperRight = statistics[xHigh][yHigh].shortStats.std;
				
				float left = lowerLeft * (1 - yRemainder) + upperLeft * yRemainder;
				float right = lowerRight * (1 - yRemainder) + upperRight * yRemainder;
				
				float std = left * (1 - xRemainder) + right * xRemainder;
				data[i][j] += random.nextGaussian() * 3 * std;
			}
			*/
	}
	
	/*
	public InverseCDF[][][] getStatistics(int levels){
		createSteerablePyramid(4, levels);
		InverseCDF stats[][][] = new InverseCDF[pyramid.bands.length][statisticsSize][statisticsSize];
		for(int i = 0; i < pyramid.bands.length; i++){
			for(int j = 0; j < statisticsSize; j++){				
				for(int k = 0; k < statisticsSize; k++){
					stats[i][j][k] = pyramid.bands[i].statistics[j][k].ICDF;					
				}
			}
		}
		return stats;
	}
	*/
	
	public void matchStatisticsOf(LocalStats targetStats){
		matchStatisticsOf(targetStats, targetStats.getNumLevels(), 4);
	}

	void saveIter(int iter){
        String iterNum = String.valueOf(iter);
        if(iterNum.length() == 1)
        	iterNum = 0 + iterNum;
        save(BatchUtil.synthesisIterDir + "iter" + iterNum + ".img");
	}
	
	public void matchStatisticsOf(LocalStats targetStats, int levels, int numIterations)
	{
		int numStatistics = ImageLoader.numStatisticTiles;
		
		//addNoise();
		
		//matchHistogram(reference, numStatistics);
	    
//		reference.createSteerablePyramid(4, levels);
		saveIter(0);
		for( int i = 0; i < numIterations; i++ )
		{
	        System.out.println("Iteration " + i);				        
	        createSteerablePyramid(4, levels);
	        //BatchUtil.compareImages(BatchUtil.targetImage, this);
			if(pyramid.bands.length != targetStats.numChannels)
				throw new RuntimeException();
			if(targetStats.numStatisticsTiles != numStatistics)
				throw new RuntimeException();
			for( int channel = 0; channel < pyramid.bands.length; channel++ )
	        {
				if(/*true/**/channel != pyramid.lowPassIndex() /* && channel != pyramid.highPassIndex()*/){
					if(targetStats.getStats()[channel].length != numStatistics || targetStats.getStats()[channel][0].length != numStatistics)
						throw new RuntimeException();
					/*
					if(i == 0)
						pyramid.bands[channel].addNoise();
					*/
					/*
					if(i == 4)
						addNoise();
					*/
					//System.out.println("   matching channel: " + channel);
					pyramid.bands[channel].matchHistogram(targetStats.getStats()[channel], numStatistics);
				}
	        }
			/*
			pyramid.createHistograms(ImageLoader.numBins, ImageLoader.numStatisticTiles);
			BatchUtil.compareImages(BatchUtil.targetImage, this);
			*/
			
			pyramid.collapse(new CircularEdgeHandler());
	        copyFrom(pyramid.collapsedImage);
	        saveIter(i + 1);
	        //matchHistogram(reference, numStatistics);
		}
		
	}
	
	
	public void matchStatisticsOf(Image reference, int levels, int numIterations)
	{
		// Generates a synthetic texture in 'Texture' that looks
		// like that presented in 'Sample'.  Initially, 'Texture'
		// contains random noise (usually), and the function 
		// modifies it by matching the histograms of the two images
		// as well as the histograms of their oriented, bandpass
		// pyramid representations using an iterative technique.
		// The histogram matching technique uses 'Bins' bins,
		// the pyramid uses 'Levels' levels and 'Orient' orientations.
		// The number of iterations performed is 'Iterations'
		matchStatisticsOf(new LocalICDF(reference, levels, ImageLoader.numStatisticTiles), levels, numIterations);
		
		/*
		int numStatistics = ImageLoader.numStatisticTiles;
		
		addNoise();
		
		//matchHistogram(reference, numStatistics);
	    
		reference.createSteerablePyramid(4, levels);
				
		for( int i = 0; i < numIterations; i++ )
		{
	        System.out.println("Iteration " + i);
			createSteerablePyramid(4, levels);
			for( int channel = 0; channel < pyramid.bands.length; channel++ )
	        {
				if(channel != pyramid.lowPassIndex())
					pyramid.bands[channel].matchHistogram(reference.pyramid.bands[channel], numStatistics);
	        }
			pyramid.collapse(new CircularEdgeHandler());
	        copyFrom(pyramid.collapsedImage);
			//matchHistogram(reference, numStatistics);
		}
		*/
	}
	
	public float sum(){
		float sum = 0;
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				sum += data[i][j];
		return sum;
	}
	
	public Image subImage(int startX, int endX, int startY, int endY){
		Image result = new Image(endX - startX, endY - startY);
		for(int i = 0; i < result.width; i++)
			for(int j = 0; j < result.height; j++)
				result.data[i][j] = data[i + startX][j + startY];
		return result;
	}
	
	public Image downSample(){
		if(width %2 == 1 || height % 2 == 1)
			throw new RuntimeException();
		Image result = new Image(width/2, height/2);
		convolve(this, Image.gaussian(2), result, true, 2, 2);
		return result;
	}
	
	void makeBorderMean(){
		float mean = avg();
		for(int i = 0; i < width; i++){
			data[i][0] = mean;
			data[i][height - 1] = mean;
		}
		for(int j = 0; j < height; j++){
			data[0][j] = mean;
			data[width - 1][j] = mean;
		}
		
	}
	
	Image laplacian;
	Image directionX;
	Image threshCurvatureBin;
	
	Vector2d[][] principleDirections;
	Vector2d[][] normalDirections;
	
	public Image edgeDetect(float thresh){
		Image result = new Image(this);				
		Image source = new Image(this);		
		directionX = new Image(width, height);
		threshCurvatureBin = new Image(width, height);
		convolve(this, Image.gaussian(2), source);
		
		Image sourceDx = new Image(width, height), sourceDy = new Image(width, height);
		convolve(source, Gx, sourceDx);
		convolve(source, Gy, sourceDy);
		Image sourceDxx = new Image(width, height), sourceDxy = new Image(width, height);
		Image sourceDyx = new Image(width, height), sourceDyy = new Image(width, height);
		convolve(sourceDx, Gx, sourceDxx);
		convolve(sourceDx, Gy, sourceDxy);
		convolve(sourceDy, Gx, sourceDyx);
		convolve(sourceDy, Gy, sourceDyy);
		principleDirections = new Vector2d[width][height];
		normalDirections = new Vector2d[width][height];
		
		Matrix hessian = new Matrix(2, 2);
		for(int i = 1; i < width - 1; i++)
			for(int j = 1; j < height - 1; j++){
				float Dxx = sourceDxx.data[i][j];
				float Dxy = .5f * (sourceDxy.data[i][j] + sourceDyx.data[i][j]);
				float Dyy = sourceDyy.data[i][j];
				hessian.set(0, 0, Dxx);
				hessian.set(1, 0, Dxy);
				hessian.set(0, 1, Dxy);
				hessian.set(1, 1, Dyy);
				EigenvalueDecomposition eig = hessian.eig();
				double[] eigenValues = eig.getRealEigenvalues();
				Matrix principleDirs = eig.getV();
				Vector2d principleDir1 = new Vector2d(principleDirs.get(0, 0), principleDirs.get(1, 0));
				Vector2d principleDir2 = new Vector2d(principleDirs.get(0, 1), principleDirs.get(1, 1));
				
				/*
				System.out.println();
				System.out.println("hessian: ");
				hessian.print(4, 1);
				System.out.println(eigenValues[0] + ", " + eigenValues[1]);
				principleDirs.print(4, 1);				
				System.out.println(principleDir1 + ", " + principleDir2);
				*/
				if(eigenValues[0] > eigenValues[1])
					System.out.println("!!!!!");
				principleDirections[i][j] = new Vector2d(principleDir1);
				normalDirections[i][j] = new Vector2d(principleDir2);
				result.data[i][j] = -(float)eig.getRealEigenvalues()[1];
				float Dx = sourceDx.data[i][j];
				float Dy = sourceDy.data[i][j];
				//directionX.data[i][j] = (float)Math.abs(Dx * principleDir2.x + Dy * principleDir2.y);
				if(directionX.data[i][j] < .00003)
					directionX.data[i][j] = 1;
				else
					directionX.data[i][j] = 0;
				
				if(result.data[i][j] < thresh)
					threshCurvatureBin.data[i][j] = 1;
				else
					threshCurvatureBin.data[i][j] = 0;
			}
	
		/*
		for(int i = 1; i < width - 1; i++)
			for(int j = 1; j < height - 1; j++){
				float left = source.data[i - 1][j];
				float right = source.data[i + 1][j];
				float up = source.data[i][j+1];
				float down = source.data[i][j-1];
				float value = source.data[i][j];
				//result.data[i][j] = data[i][j] = source.data[i][j];
				//result.data[i][j] = source.data[i][j] - .25f * (left + right + up + down);
				float yCurv = value - .5f * (up + down);
				float xCurv = value - .5f * (left + right);
				directionX.data[i][j] = Math.min(xCurv, yCurv);
				
				if(result.data[i][j] < thresh){
					if(xCurv < yCurv)
						directionX.data[i][j] = 1;
					else
						directionY.data[i][j] = 1;
				}
				
				//result.data[i][j] = (float)Math.pow(result.data[i][j], 3);
				
				if(result.data[i][j] < thresh)
					result.data[i][j] = 1;
				else
					result.data[i][j] = 0;
				
				//result.data[i][j] = (float)Math.pow(result.data[i][j], 2);
 
			}
		directionX.makeBorderMean();
		result.makeBorderMean();
		*/
		/*
		Image result2 = new Image(width, height);
		
		for(int i = 1; i < width - 1; i++)
			for(int j = 1; j < height - 1; j++){
				int lowerCount = 0;
				float value = data[i][j];
				float left = data[i - 1][j];
				float right = data[i + 1][j];
				float up = data[i][j+1];
				float down = data[i][j-1];
				if(left < value) lowerCount++;
				if(right < value) lowerCount++;
				if(up < value) lowerCount++;
				if(down < value) lowerCount++;
				
				if(lowerCount >= 2)
					result2.data[i][j] = 0;
				else if(result.data[i][j] > 0)
					result2.data[i][j] = 1;
			}
		*/
		return result;
		
	}
	
	static void convolve(Image I, Image filter, Image result){
		Image.convolve(I, filter, result, false, 1, 1);
	}
	
	public Image cannyEdgeDetect(){
		Image gaussian = gaussian(2);
		Image smoothed = new Image(width, height);
		Image.convolve(this, gaussian, smoothed);
		Image xDerivative = new Image(width, height);
		Image yDerivative = new Image(width, height);
				
		Image.convolve(smoothed, Gx, xDerivative);
		Image.convolve(smoothed, Gy, yDerivative);
		
		Image magDerivative = new Image(width, height);
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				magDerivative.data[i][j] = (float)Math.sqrt(xDerivative.data[i][j] * xDerivative.data[i][j] + 
													 yDerivative.data[i][j] * yDerivative.data[i][j]);
		
		//int[][][]  
		
		return magDerivative;
	}
	/*
	public Image rain(){
		Random rand = new Random();
		Image result = new Image(width, height);
		for(int iter = 0; iter < width * height; iter++){
			int x = rand.nextInt(width - 2) + 1;
			int y = rand.nextInt(height -2) + 1;
			while(true){
				if(x == 0 || y == 0 || x == width - 1 || y == width - 1)
					break;			
				float value = data[x][y];
				float minValue = Float.POSITIVE_INFINITY;
				
			
			
		}
	}
	*/
	
	static Image gaussian(float sigma){
		int size = (int)(sigma * 2 * 2);
		Image result = new Image(size, size);
		int center = size/2;
		for(int i = 0; i < size; i++){
			for(int j = 0; j < size; j++){
				float distSq = (i - center) * (i - center) + (j - center) * (j - center);
				result.data[i][j] = (float)(1.0 / Math.sqrt(2 * Math.PI) * Math.exp(-distSq / 2 / sigma));
			}		
		}
		return result;
	}
	
	static Image Gx;
	
	
	static Image Gy;
	
	static{
		Gx = new Image(3, 3);
		Gx.data[0][0] = -1;		Gx.data[2][0] = 1;
		Gx.data[0][1] = -2;		Gx.data[2][1] = 2;
		Gx.data[0][2] = -1;		Gx.data[2][2] = 1;
		//Gx.normalize();
		Gx.scale(1.0f/8);
		
		Gy = new Image(3, 3);
		Gy.data[0][0] = -1;		Gy.data[0][2] = 1;
		Gy.data[1][0] = -2;		Gy.data[1][2] = 2;
		Gy.data[2][0] = -1;		Gy.data[2][2] = 1;
		Gy.scale(1.0f/8);
		//Gy.normalize();
		
		//System.out.println("Gx:" + Gx + "\n");
		//System.out.println("Gy:" + Gy + "\n");
	}
	
	
	static Pixel[] neighborDirs = new Pixel[]{new Pixel(1, 0), new Pixel(1, 1), new Pixel(0, 1),
									new Pixel(-1, 1), new Pixel(-1, 0), new Pixel(-1, -1),
									new Pixel(0, -1), new Pixel(1, -1)};
	
	static Pixel[] nextPixels(int x, int y, Vector2d dir){
		double angle = Math.atan2(dir.y, dir.x);
		if(angle < 0) angle += 2 * Math.PI;
		//angle += Math.PI; //zero to 2PI
		angle += Math.PI / 8;
		int neighborIndex = (int)(angle / 2 / Math.PI * 8) % neighborDirs.length;
		//System.out.println("  " + dir + " , "  + angle + ", " + neighborIndex);
		Pixel[] pixels = new Pixel[3];
		for(int i = 0; i < pixels.length; i++){			
			Pixel pixelDir = neighborDirs[(neighborIndex + i - pixels.length / 2 + neighborDirs.length) % neighborDirs.length];
			pixels[i] = new Pixel(x + pixelDir.x, y + pixelDir.y);			
		}
		return pixels;
	}
	
	static Pixel nextPixel(int x, int y, Vector2d dir){
		double angle = Math.atan2(dir.y, dir.x);
		if(angle < 0) angle += 2 * Math.PI;
		//angle += Math.PI; //zero to 2PI
		angle += Math.PI / 8;
		int neighborIndex = (int)(angle / 2 / Math.PI * 8) % neighborDirs.length;
		//System.out.println("  " + dir + " , "  + angle + ", " + neighborIndex);
		Pixel pixelDir = neighborDirs[neighborIndex];
		return new Pixel(x + pixelDir.x, y + pixelDir.y);
	}

	
	static float getValue(Image image, Pixel pixel){
		return image.data[pixel.x][pixel.y];
	}
	
	static void followWrinkle(int startX, int startY, Vector2d startDir,
			Image curvatureImage, Vector2d[][] directions, Image result){

		int x = startX, y = startY;		
		Vector2d avgDir = new Vector2d(startDir);
		while(true){
			result.data[x][y] = 1;
			Vector2d dir = directions[x][y];
			avgDir.scaleAdd(.3, dir, avgDir);
			avgDir.normalize();
			
			Pixel[] nextPixels = nextPixels(x, y, avgDir);
			Pixel nextPixel = nextPixels[0];
			float minCurvature = getValue(curvatureImage, nextPixel);
			for(int i = 1; i < nextPixels.length; i++){
				float curvature = getValue(curvatureImage, nextPixels[i]);
				if(curvature < minCurvature){
					minCurvature = curvature;
					nextPixel = nextPixels[i];
				}
			}
			
			x = nextPixel.x;
			y = nextPixel.y;
			if(x <= 0 || x >= curvatureImage.width-1 ||
				y <= 0 || y >= curvatureImage.height - 1)
				break;
			
			if(curvatureImage.data[x][y] < -0.000008f){
				
			}
			else{
				break;
			}
		}
		
	}
	
	static Image getRegion2(int startX, int startY, Image curvatureImage, Vector2d[][] directions){
		Image result = new Image(curvatureImage.width, curvatureImage.height);
		Vector2d startDir = new Vector2d(directions[startX][startY]);
		followWrinkle(startX, startY, startDir, curvatureImage, directions, result);
		startDir.negate();
		followWrinkle(startX, startY, startDir, curvatureImage, directions, result);
		return result;
	}
	
	Image getRegion(int startX, int startY, Vector2d[][] directions){
		boolean visited[][] = new boolean[width][height];
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				visited[i][j] = false;
		
		LinkedList toVisit = new LinkedList();
		toVisit.add(new Pixel(startX, startY));
		Vector2d dirSum = new Vector2d(directions[startX][startY]);
		Image result = new Image(width, height);

		while(!toVisit.isEmpty()){
			Pixel pixel = (Pixel)toVisit.getFirst();
			toVisit.removeFirst();
			int x = pixel.x, y = pixel.y;
			System.out.println(toVisit.size() + "; " + x + ", " + y);
			/*
			if(x < 0 || x >= width || y < 0 || y >= width)
				continue;
			*/
			visited[x][y] = true;
			for(int i = -1; i <= 1; i++){
				for(int j = -1; j <= 1; j++){
					if(!(i == 0 && j == 0)){
						int nextX = x + i;
						int nextY = y + j;
						if(!(nextX < 0 || nextX >= width || nextY < 0 || nextY >= width) &&
								!visited[nextX][nextY]){
							if(data[nextX][nextY] > 0){
								Vector2d dir = directions[nextX][nextY];
								if(dir.dot(dirSum) < 0)
									dir.negate();
								double cosAngle = dir.dot(dirSum) / dirSum.length();
								visited[nextX][nextY] = true;
								if(cosAngle > .7/*.86*/){								
									toVisit.addLast(new Pixel(nextX, nextY));
									result.data[nextX][nextY] = 1;
								}
							}
						}							
					}
				}
			}			
		}	
		
		return result;
	}
	
	static long followWrinkleTime = 0;
	
	static List followWrinkle2(int startX, int startY, Vector2d startDir,
			Image binImage, Vector2d[][] directions){
		long followWrinkleStartTime = System.currentTimeMillis();
		List result = new ArrayList(); 
		int x = startX, y = startY;		
		Vector2d avgDir = new Vector2d(startDir);
		while(true){
			if(binImage.data[x][y] < 1)
				System.out.println("error!");
			result.add(new Pixel(x, y));
			//result.data[x][y] = 1;
			Vector2d dir = new Vector2d(directions[x][y]);
			if(dir.dot(avgDir) < 0)
				dir.negate();
			avgDir.scaleAdd(.2, dir, avgDir);
			avgDir.normalize();
			
			Pixel[] nextPixels = nextPixels(x, y, avgDir);
			Pixel bestPixel = null;
			double maxCosAngle = 0;
			for(int i = 0; i < nextPixels.length; i++){
				Pixel nextPixel = nextPixels[i];
				if( getValue(binImage, nextPixel) > 0){
					Vector2d nextDir = new Vector2d(directions[nextPixel.x][nextPixel.y]);
					if(nextDir.dot(avgDir) < 0)
						nextDir.negate();
					double cosAngle = nextDir.dot(avgDir);
					if(cosAngle > maxCosAngle){
						maxCosAngle = cosAngle;
						bestPixel = nextPixel;
					}					
				}
			}
			if(bestPixel == null || maxCosAngle < .8) break;
			
			x = bestPixel.x;
			y = bestPixel.y;
			if(x <= 0 || x >= binImage.width-1 ||
				y <= 0 || y >= binImage.height - 1)
				break;
			
		}
		followWrinkleTime += System.currentTimeMillis()  - followWrinkleStartTime;
		return result;
	}
	
	static long getRegionTime = 0;
	
	static List getRegion3(int startX, int startY, Image binImage, Vector2d[][] directions){
		long getRegionStartTime = System.currentTimeMillis();
		//System.out.println("Getting region: " + startX + " " + startY);
		//Image result = new Image(binImage.width, binImage.height);
		Vector2d startDir = new Vector2d(directions[startX][startY]);
		List wrinkles = followWrinkle2(startX, startY, startDir, binImage, directions);
		wrinkles.remove(0);
		int numPixels = wrinkles.size();
		for(int i = 0; i < numPixels/2; i++){
			Object temp = wrinkles.get(i);
			wrinkles.set(i, wrinkles.get(numPixels - i - 1));
			wrinkles.set(numPixels - i - 1, temp);
		}
		startDir.negate();
		wrinkles.addAll(followWrinkle2(startX, startY, startDir, binImage, directions));
		//System.out.println("region: " + wrinkles);
		getRegionTime += System.currentTimeMillis() - getRegionStartTime;
		return wrinkles;
	}

	static Image getRegion4(int startX, int startY, Image binImage, Vector2d[][] directions){
		Image result = new Image(binImage.width, binImage.height);
		Vector2d startDir = new Vector2d(directions[startX][startY]);
		int width = binImage.width;
		int height = binImage.height;
		for(int i = 1; i < width - 1; i++){
			for(int j = 1; j < height - 1; j++){
				if(binImage.data[i][j] > 0){
					Vector2d dir = directions[i][j];
					if(startDir.dot(dir) < 0)
						dir.negate();
					double angle = dir.angle(startDir);
					if(angle < Math.PI / 6)
						result.data[i][j] = 1;
					else
						result.data[i][j] = 0;
				}
			}
		}
		
		return result;
	}
	
	
	void scale(float scale){
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				data[i][j] *= scale;		
	}
	
	void normalize(){
		float sum = sum();
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				data[i][j] /= sum;
	}
	
	void suppress(Image curvature, Vector2d[][] normalDirs){
		System.out.println("Suppressing");
		for(int i = 1; i < width; i++){
			for(int j = 1; j < height; j++){
				if(data[i][j] > 0){
					float value = curvature.data[i][j];
					float up = getValue(curvature, nextPixel(i, j, normalDirs[i][j]));
					normalDirs[i][j].negate();
					float down = getValue(curvature, nextPixel(i, j, normalDirs[i][j]));
					if(up < value || down < value)
						data[i][j] = 0;
				}
			}
		}
		createDrawnImage(false);
	}
	
	void drawPixels(Set pixels, float value){
		for(Iterator iter = pixels.iterator(); iter.hasNext();){
			Pixel pixel = (Pixel)iter.next();
			data[pixel.x][pixel.y] = value;
		}
	}
	
	static long removeWrinkleTime = 0;
	
	static List removeWrinkles(Image binImage, Image thinnedBinImage, Image heightmapImage, Vector2d[][] directions, Vector2d[][] normalDirs){
		Image fromBinImage = new Image(thinnedBinImage);
		Image diffImage = new Image(heightmapImage);
		diffImage.raise(5);
		diffImage.subtract(heightmapImage);
		
		System.out.println("Extracting wrinkles...");
		List wrinkles = new ArrayList();
		
		List onPixelsList = new ArrayList();
		for(int i = 0; i < fromBinImage.width; i++){
			for(int j = 0; j < fromBinImage.height; j++){
				if(fromBinImage.data[i][j] > 0){
					onPixelsList.add(new Pixel(i, j, heightmapImage.data[i][j]));
				}
			}
		}
		Object[] onPixels = onPixelsList.toArray();
		Arrays.sort(onPixels);
				
		//while(true){
		long removeWrinkleStartTime = System.currentTimeMillis();
		for(int i = 0; i < onPixels.length; i++){
		//for(int iter = 0; iter < 1000; iter++){
			/*
			int minX = -1, minY = -1;
			float minCurvature = Float.POSITIVE_INFINITY;
			for(int i = 3; i < fromBinImage.width - 3; i++){
				for(int j = 3; j < fromBinImage.height - 3; j++){
					if(fromBinImage.data[i][j] > 0){
						if(heightmapImage.data[i][j] < minCurvature){
							minCurvature = heightmapImage.data[i][j];
							minX = i;
							minY = j;
						}
					}
				}
			}
			if(minX == -1)
				break;
			*/
			if(i % 20000 == 0)
				System.out.println("  " + i + " / " + onPixels.length);
			Pixel pixel = (Pixel)onPixels[i];
			//System.out.println("  " + fromBinImage.sum() + "; " + heightmapImage.data[pixel.x][pixel.y]);
			if(fromBinImage.data[pixel.x][pixel.y] > 0){
				int minX = pixel.x;
				int minY = pixel.y;
				List pixels = getRegion3(minX, minY, fromBinImage, directions);
				//System.out.println(pixels.size());				
				if(pixels.size() > 20){
					wrinkles.add(new Wrinkle(pixels, diffImage, pixel));
					/*
					if(wrinkles.size() % 50 == 0)
						System.out.println("  wrinkle: " + wrinkles.size());
					*/
				}
				//result.drawPixels(linePixels, iter);
				fromBinImage.subtractPixels(pixels);
				//fromBinImage.createDrawnImage(false);
			}
		}
		
		removeWrinkleTime = System.currentTimeMillis() - removeWrinkleStartTime;
		System.out.println("Remove wrinkle time: " + removeWrinkleTime);
		System.out.println("Region time: " + getRegionTime);
		System.out.println("Follow wrinkle time: " + followWrinkleTime);
		
		for(int i = 0; i < wrinkles.size(); i++){
			Wrinkle wrinkle = (Wrinkle)wrinkles.get(i);
			wrinkle.calculateWidth(binImage, normalDirs);
			//System.out.println("  " + wrinkle.avgWidth);
		}
			
		return wrinkles;
	}			
	
	void subtract(Image image){
		if(image.height != height || image.width != width)
			throw new RuntimeException();
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				data[i][j] -= image.data[i][j];
	}

	void add(Image image){
		if(image.height != height || image.width != width)
			throw new RuntimeException();
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				data[i][j] += image.data[i][j];
	}

	void minSubtract(Image image){
		if(image.height != height || image.width != width)
			throw new RuntimeException();
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				data[i][j] = Math.min(data[i][j], -image.data[i][j]);
	}

	void subtractBinary(Image image){
		if(image.height != height || image.width != width)
			throw new RuntimeException();
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				if(image.data[i][j] > 0)
					data[i][j] = 0;
			
	}
	
	void subtractPixels(List pixels){
		for(Iterator iter = pixels.iterator(); iter.hasNext();){
			Pixel pixel = (Pixel)iter.next();
			data[pixel.x][pixel.y] = 0;
		}		
	}

	void addPixels(List pixels){
		for(Iterator iter = pixels.iterator(); iter.hasNext();){
			Pixel pixel = (Pixel)iter.next();
			data[pixel.x][pixel.y] = 1;
		}		
	}
	
	void raise(int range){		
		float[][] newData = new float[width][height];
		for(int i = range; i < width - range; i++){
			for(int j = range; j < height- range; j++){
				float max = Float.NEGATIVE_INFINITY;
				for(int dx = -range; dx <= range; dx++){
					for(int dy = -range; dy <= range; dy++){
						int x = i + dx;
						int y = j + dy;
						if(data[x][y] > max)
							max = data[x][y];
					}
				}
				newData[i][j] = max;
			}
		}
		data = newData;
						
	}

	void addWrinklePatchDisplacement(WrinklePatch wrinklePatch){
		for(int i = 0; i < wrinklePatch.pixels.size(); i++){
			Pixel pixel = (Pixel)wrinklePatch.pixels.get(i);
			int x = pixel.x;
			int y = pixel.y;
			data[x][y] = Math.max(pixel.height, data[x][y]);
		}
	
	}

	
	void addWrinkleDisplacement(Wrinkle wrinkle){
		/*
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++)
				data[i][j] = 0;
		*/
		
		int maxWidthInt = (int)Math.ceil(wrinkle.maxWidth);
		float maxWidthSq = wrinkle.maxWidth * wrinkle.maxWidth;
		for(int i = 0; i < wrinkle.pixels.size(); i++){
			Pixel pixel = (Pixel)wrinkle.pixels.get(i);
			int x0 = pixel.x, y0 = pixel.y;
			for(int dx = -maxWidthInt; dx <= maxWidthInt; dx++){				
				for(int dy = -maxWidthInt; dy <= maxWidthInt; dy++){					
					float distSq = dx*dx + dy*dy;
					int x = x0 + dx;
					int y = y0 + dy;
					if(x >= 0 && x < width && y >= 0 && y < height){
						if(distSq < pixel.width * pixel.width){												
							float dist = (float)Math.sqrt(distSq);
							float height = pixel.height - dist * pixel.height / pixel.width;
							//System.out.println(height + "; " + dist + "; " + pixel.height);
							data[x][y] = Math.max(height, data[x][y]);
						}
					}
				}
			}
		}
				
	}
	
	void applyWrinkles(List wrinkles){
		Image wrinkleDisplacement = new Image(width, height);
		
		for(int i = 0; i < wrinkles.size(); i++){
			if(i % 200 == 0)
				System.out.println("  applying wrinkle: " + i);
			Wrinkle wrinkle = (Wrinkle)wrinkles.get(i);
			wrinkleDisplacement.addWrinkleDisplacement(wrinkle);
			//minSubtract(wrinkleDisplacement);
			//System.out.println("min: " + min() + "; max of displacement: " + wrinkleDisplacement.max());
			/*
			for(int j = 0; j < wrinkle.pixels.size(); j++){
				Pixel pixel = (Pixel)wrinkle.pixels.get(j);
				data[pixel.x][pixel.y] -= pixel.height;
			}*/
		}
		subtract(wrinkleDisplacement);
		createDrawnImage(true);
	}
	
	void applyWrinklePatches(List wrinklePatches){
		Image wrinkleDisplacement = new Image(width, height);
		
		for(int i = 0; i < wrinklePatches.size(); i++){
			if(i % 200 == 0)
				System.out.println("  applying wrinkle patch: " + i);
			WrinklePatch wrinklePatch = (WrinklePatch)wrinklePatches.get(i);
			wrinkleDisplacement.addWrinklePatchDisplacement(wrinklePatch);
			//minSubtract(wrinkleDisplacement);
			//System.out.println("min: " + min() + "; max of displacement: " + wrinkleDisplacement.max());
			/*
			for(int j = 0; j < wrinkle.pixels.size(); j++){
				Pixel pixel = (Pixel)wrinkle.pixels.get(j);
				data[pixel.x][pixel.y] -= pixel.height;
			}*/
		}
		subtract(wrinkleDisplacement);
		createDrawnImage(true);
	}

	
	float getLength(Pixel from, Vector2d dir){
		int length = 0;
		
		float dt_x = Math.abs(1/(float)dir.x), dt_y = Math.abs(1/(float)dir.y);
		float t_next_x = dt_x / 2, t_next_y = dt_y / 2;
		int i = from.x, j = from.y;
		int sign_x = 1;
		if(dir.x < 0) sign_x = -1;
		int sign_y = 1;
		if(dir.y < 0) sign_y = -1;
		
		while(data[i][j] > 0){
			if(i < 0 || i >= width || j < 0 || j >= height)
				break;
			//System.out.println("  (" + i + ", " + j + ")");
			length++;
			if(t_next_x < t_next_y){
				i += sign_x;
				t_next_x += dt_x;
			}
			else{
				j += sign_y;
				t_next_y += dt_y;
			}
		}
		return (float)Math.sqrt((i - from.x) * (i - from.x) + (j - from.y)*(j - from.y));
	}
	
	
	public static Image avgOf(Image[] inputs){
		int height = inputs[0].height;
		int width = inputs[0].width;
		for(int i = 0; i < inputs.length; i++){
			if(inputs[i].width != width || inputs[i].height != height)
				throw new RuntimeException();
		}
		
		Image output = new Image(width, height);
		for(int i = 0; i < inputs.length; i++)
			output.add(inputs[i]);
		
		output.scale(1.0f / inputs.length);
		return output;
	}
		
	public void addPerlinNoise(float scale){
		Image noiseImage = perlinNoise(width, height);
		noiseImage.scale(scale);
		add(noiseImage);
	}
	
	public static Image perlinNoise(int width, int height){
		Image image = new Image(width, height);
		for(int i = 0; i < width; i++)
			for(int j = 0; j < height; j++){
				image.data[i][j] = (float)PerlinNoise.noise(((float)i) / 6, ((float)j) / 6, 0);
				//System.out.println(image.data[i][j]);
			}
		
		System.out.println(image.min() + " / " + image.max());
		return image;
	}

	public static void main(String[] args){
		Image image = new Image("base_wojciech.img");
		image.saveAs("test.bmp", true);
	}
	
}
