package jogl;

import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;

public class LocalShortStats extends LocalStats{
	final static float minSTD = 0.00000001f; 

	ShortStats[][][] shortStats;
	
	public LocalShortStats(){
		//this.numChannels = ImageLoader.numc
		this.numStatisticsTiles = ImageLoader.numStatisticTiles;
	}
	
	public LocalShortStats(String file){
		loadFrom(file);
	}
	
	public LocalShortStats(Image image, int levels){
		this();
		image.createSteerablePyramid(4, levels);
		numChannels = image.pyramid.bands.length;
		shortStats = new ShortStats[numChannels][numStatisticsTiles][numStatisticsTiles];
		for(int channel = 0; channel < numChannels; channel++)
			for(int i = 0; i < numStatisticsTiles; i++)
				for(int j = 0; j < numStatisticsTiles; j++)
					shortStats[channel][i][j] = image.pyramid.bands[channel].statistics[i][j].shortStats;
	}
	
	Object[][][] getStats(){
		return shortStats;
	}
	
	void saveAs(String file){
		try{
			System.out.println("Saving short stats to " + file);
			System.out.println("  ch: " + numChannels + " tiles: " + 
					numStatisticsTiles);				
			
			FileOutputStream fstream = new FileOutputStream(file);
			BufferedOutputStream bstream = new BufferedOutputStream(fstream, 
					numChannels * numStatisticsTiles * numStatisticsTiles * 4 + 8);// vertices.size()*6*4 + faces.size() * 13);
			DataOutputStream dstream = new DataOutputStream(bstream);				
	
			
			//write description
			dstream.writeInt(numChannels);
			dstream.writeInt(numStatisticsTiles);
			
			
			for(int i = 0; i < numChannels; i++)
				for(int j = 0; j < numStatisticsTiles; j++)
					for(int k = 0; k < numStatisticsTiles; k++){
						dstream.writeFloat(shortStats[i][j][k].mean);
						dstream.writeFloat(shortStats[i][j][k].std);
					}
			dstream.close();
		}		
		catch(Exception e){
			e.printStackTrace();
		}
	}
	
	void loadFrom(String file){
		try{
			int fileLength = (int)(new File(file)).length();
			byte fileBytes[] = new byte[fileLength];
			FileInputStream fStream = new FileInputStream(file);
			fStream.read(fileBytes);
			fStream.close();			
			System.out.println("Read file " + file + ", processing..");
			ByteArrayInputStream bStream = new ByteArrayInputStream(fileBytes);
			DataInputStream dstream = new DataInputStream(bStream);
			
			numChannels = dstream.readInt();
			numStatisticsTiles = dstream.readInt();
			
			shortStats = new ShortStats[numChannels][numStatisticsTiles][numStatisticsTiles];
			System.out.println("  ch: " + numChannels + " tiles: " + 
					numStatisticsTiles);				

			
			int counter = 0; 

			for(int i = 0; i < numChannels; i++)
				for(int j = 0; j < numStatisticsTiles; j++)
					for(int k = 0; k < numStatisticsTiles; k++){							
						float mean = dstream.readFloat();
						float std = dstream.readFloat();
						
						if(std <= minSTD){
							//System.out.println(std);
							counter++;
							std = minSTD;
							//throw new RuntimeException();
						}

						shortStats[i][j][k] = new ShortStats(mean, std);
					}
			System.out.println("   adjusted stds: " + counter);
			dstream.close();
		}
		catch(Exception e){
			e.printStackTrace();
		}		
	}

	public boolean equals(Object o){
		if(!(o instanceof LocalShortStats))
			return false;
		else{
			LocalShortStats stats = (LocalShortStats)o;
			if(stats.numChannels != numChannels)
				return false;
			if(stats.numStatisticsTiles != numStatisticsTiles)
				return false;
			
			for(int i = 0; i < numChannels; i++)
				for(int j = 0; j < numStatisticsTiles; j++)
					for(int k = 0; k < numStatisticsTiles; k++)											
						if(shortStats[i][j][k].mean != stats.shortStats[i][j][k].mean ||
								shortStats[i][j][k].std != stats.shortStats[i][j][k].std)
							return false;
			return true;
		}
	}
	
	public static void main(String[] args){
		Image image = new Image(BatchUtil.outputDir + "temp\\subject-30k.img");
		LocalShortStats stats = new LocalShortStats(image, 4);
		String file = BatchUtil.outputDir + "temp\\testStats.ss"; 
		stats.saveAs(file);
		LocalShortStats newStats = new LocalShortStats(file);
		System.out.println(newStats.equals(stats));
	}

}
