/*
 * Created on Aug 31, 2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jogl;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import javax.vecmath.Vector2d;
import javax.vecmath.Vector2f;
import javax.vecmath.Vector3d;

import net.java.games.jogl.GL;

/**
 * @author golovins
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class WrinkleModel {
	static Random rand = new Random(0);
	
	Vector2f center;
	float length;
	float angle;
	
	boolean farSide = false;
	
	Vector2f first, last;
	
	Wrinkle originalWrinkle;
	
	public WrinkleModel(WrinkleModel wrinkle){
		this.center = new Vector2f(wrinkle.center);
		this.length = wrinkle.length;
		this.angle = wrinkle.angle;
		this.originalWrinkle = wrinkle.originalWrinkle;
	}

	public WrinkleModel(float startX, float startY, float endX, float endY){
		set(startX, startY, endX, endY);
	}
	
	public WrinkleModel(){
		center = new Vector2f();
	}
	
	public WrinkleModel(Wrinkle wrinkle){
		Pixel firstPixel = (Pixel)wrinkle.pixels.get(0);
		Pixel lastPixel = (Pixel)wrinkle.pixels.get(wrinkle.pixels.size() - 1);
		set(firstPixel.x, firstPixel.y, lastPixel.x, lastPixel.y);
		originalWrinkle = wrinkle;
	}
	
	public void set(float startX, float startY, float endX, float endY){
		first = new Vector2f(startX, startY);
		last = new Vector2f(endX, endY);
		
		center = new Vector2f();
		center.interpolate(first, last, .5f);
		Vector2f diff = new Vector2f();
		diff.sub(last, first);
		length = diff.length();
		angle = getAngle(diff.x, diff.y);
		/*
		if(center.x >= ImageLoader.imageWidth / 2){
			farSide = true;
			center.x = ImageLoader.imageWidth - center.x - 1;
		}
		*/
	}

	public float centerDistanceTo(WrinkleModel wModel){
		return (float)Math.sqrt((center.x - wModel.center.x) * (center.x - wModel.center.x) + 
				 (center.y - wModel.center.y) * (center.y - wModel.center.y));
		
	}
	
	public float orientationDistanceTo(WrinkleModel wModel){
		float otherAngle = wModel.angle;
		float diff = Math.abs(otherAngle - angle);
		if(diff > Math.PI / 2)
			diff = (float)Math.PI - diff;
		return diff;
	}
	
	public float distanceTo(WrinkleModel wModel){
						 //(length - wModel.length) * (length - wModel.length)*/);
		return centerDistanceTo(wModel) + 150 * orientationDistanceTo(wModel);
	}
	
	static float getAngle(double x, double y){
		float angle = (float)Math.atan2(y, x);	
		if(angle < 0)
			angle += Math.PI;		
		return angle;
	}
	
	static WrinkleModel getCentroid(List wrinkles){
		WrinkleModel avg = new WrinkleModel();
		Vector2d avgDir = new Vector2d();
		for(int i = 0; i < wrinkles.size(); i++){
			WrinkleModel wrinkle = (WrinkleModel)wrinkles.get(i);
			avg.center.add(wrinkle.center);
			avg.length += wrinkle.length;
			if(i == 0){
				avgDir.set(Math.cos(wrinkle.angle), Math.sin(wrinkle.angle));
			}
			else{
				Vector2d nextDir = new Vector2d(Math.cos(wrinkle.angle), Math.sin(wrinkle.angle));
				if(nextDir.dot(avgDir) < 0)
					nextDir.negate();
				avgDir.add(nextDir);
			}
			//System.out.println("  " + avgDir);
		}
		avg.center.scale(1.0f / wrinkles.size());
		avg.length /= wrinkles.size();
		avg.angle = getAngle(avgDir.x, avgDir.y);
		return avg;
		
	}

	//static float minDist()
	
	static float closestDist(WrinkleModel from, WrinkleModel[] wrinkles){
		float closestDist = Float.POSITIVE_INFINITY;
		//System.out.println("from: " + from);
		for(int i = 0; i < wrinkles.length; i++){
			if(wrinkles[i] == null)
				continue;
			float dist = from.distanceTo(wrinkles[i]);
			//System.out.println(wrinkles[i] + ": " + dist);
			if(dist < closestDist)
				closestDist = dist;
		}
		//System.out.println("closest: " + closestDist);
		return closestDist;		
	}
	
	static WrinkleModel[][] cluster(WrinkleModel wrinkles[], int numClusters){
		
		WrinkleModel centroids[] = new WrinkleModel[numClusters]; 
		
		for(int i = 0; i < centroids.length; i++){
			int index = rand.nextInt(wrinkles.length);
			while(closestDist(wrinkles[index], centroids) < 10){
				index = rand.nextInt(wrinkles.length);
			}
			centroids[i] = new WrinkleModel(wrinkles[index]);
		}
		
		int[] cluster = new int[wrinkles.length];
		for(int i = 0; i < cluster.length; i++)
			cluster[i] = -1;
		
		List clusterMembers[] = new List[numClusters];
		
		boolean changed = true;
		int iteration = 0;
		while(changed){		
			System.out.println(" k-means iteration: " + iteration);
			iteration++;
			/*
			System.out.println("centroids: ");
			for(int i = 0; i < centroids.length; i++)
				System.out.println("  " + centroids[i]);
			*/
			changed = false;

			for(int i = 0; i < clusterMembers.length; i++)
				clusterMembers[i] = new ArrayList();
			
			for(int i = 0; i < wrinkles.length; i++){
				float closestDist = Float.POSITIVE_INFINITY;
				int closestCluster = -2;
				for(int j = 0; j < centroids.length; j++){
					float dist = wrinkles[i].distanceTo(centroids[j]);
					if(dist < closestDist){
						closestDist = dist;
						closestCluster = j;
					}
				}
				clusterMembers[closestCluster].add(wrinkles[i]);
				
				if(cluster[i] != closestCluster){
					cluster[i] = closestCluster;
					changed = true;
				}
			}
			/*
			System.out.println("clusters: ");
			for(int i = 0; i < numClusters; i++){
				System.out.println("  cluster " + i);
				for(int j = 0; j < clusterMembers[i].size(); j++)
					System.out.println("    " + clusterMembers[i].get(j));
			}
			*/
			
			
			for(int i = 0; i < numClusters; i++){
				centroids[i] = WrinkleModel.getCentroid(clusterMembers[i]);
			}
		}
		
		WrinkleModel clusteredWrinkles[][] = new WrinkleModel[numClusters][];
		for(int i = 0; i < numClusters; i++){
			clusteredWrinkles[i] = new WrinkleModel[clusterMembers[i].size()];
			for(int j = 0; j < clusteredWrinkles[i].length; j++)
				clusteredWrinkles[i][j] = (WrinkleModel)clusterMembers[i].get(j);
		}
		return clusteredWrinkles;
	}
	
	public String toString(){
		return center + ", " + length + "; " + angle / Math.PI * 180;
	}
	
	
	public static WrinkleCluster[] getWrinkleClusters(List wrinkles, int numClusters){
		WrinkleModel wrinkleModels[] = new WrinkleModel[wrinkles.size()];
		for(int i = 0; i < wrinkleModels.length; i++)
			wrinkleModels[i] = new WrinkleModel((Wrinkle)wrinkles.get(i));
		
		/*
		System.out.println("Wrinkles: ");
		for(int i = 0; i < wrinkleModels.length; i++)
			System.out.println(wrinkleModels[i]);
		*/
		
		WrinkleModel clusters[][] = cluster(wrinkleModels, numClusters);
		
		/*
		System.out.println("\n\nClusters: ");
		for(int i = 0; i < clusters.length; i++){
			System.out.println("  cluster " + i);
			for(int j = 0; j < clusters[i].length; j++)
				System.out.println("     " + clusters[i][j]);					
		}
		*/

		WrinkleCluster[] wrinkleClusters = new WrinkleCluster[clusters.length];
		for(int i = 0; i < wrinkleClusters.length; i++){
			wrinkleClusters[i] = new WrinkleCluster(clusters[i]);
		}
		
		return wrinkleClusters;
	}
	
	public static void main(String[] args) {		
		/*
		WrinkleModel wrinkles[] = new WrinkleModel[3];
		wrinkles[0] = new WrinkleModel(0, 0, 1, 0);
		wrinkles[1] = new WrinkleModel(0, 0, -1, 0);
		wrinkles[2] = new WrinkleModel(0, 0, 0, 1);
		//wrinkles[2] = new WrinkleModel(0, 0, 1, .00001f);
		
		System.out.println("Wrinkles: ");
		for(int i = 0; i < wrinkles.length; i++)
			System.out.println(wrinkles[i]);

		System.out.println();
		for(int i = 0; i < wrinkles.length; i++){
			for(int j = 0; j < wrinkles.length; j++)
				System.out.print(wrinkles[i].orientationDistanceTo(wrinkles[j]) + " ");
			System.out.println();
		}
		System.out.println();
		
		List wrinkleList = new ArrayList();
		for(int i = 0; i < wrinkles.length; i++)
			wrinkleList.add(wrinkles[i]);
		System.out.println("Centroid: " + getCentroid(wrinkleList));
		*/
		/*
		WrinkleModel centroids[] = new WrinkleModel[4];
		centroids[0] = new WrinkleModel(0, 0, 0, 1); 
		centroids[1] = new WrinkleModel(5, 0, 5, 1); 
		centroids[2] = new WrinkleModel(5, 5, 5, 6); 
		centroids[3] = new WrinkleModel(0, 5, 0, 6);
		
		WrinkleModel wrinkles[] = new WrinkleModel[22];
		for(int i = 0; i < wrinkles.length; i++){			
			int centroidIndex = rand.nextInt(centroids.length);
			System.out.println(centroidIndex);
			WrinkleModel centroid = centroids[centroidIndex];
			wrinkles[i] = new WrinkleModel(centroid);
			wrinkles[i].center.x += rand.nextFloat() * 4;			
		}
		
		System.out.println("Wrinkles: ");
		for(int i = 0; i < wrinkles.length; i++)
			System.out.println(wrinkles[i]);
		
		WrinkleModel clusters[][] = cluster(wrinkles, centroids.length);
		
		System.out.println("\n\nClusters: ");
		for(int i = 0; i < clusters.length; i++){
			System.out.println("  cluster " + i);
			for(int j = 0; j < clusters[i].length; j++)
				System.out.println("     " + clusters[i][j]);					
		}*/
	}

	public void draw(GL gl, Vector3d color){
		gl.glColor3d(color.x, color.y, color.z);
		gl.glBegin(GL.GL_LINES);
		gl.glVertex2f(first.x, first.y);
		gl.glVertex2f(last.x, last.y);
		gl.glEnd();				
	}

	
}
