package jogl;

import java.util.*;

import javax.vecmath.Vector3d;

import net.java.games.jogl.GL;

public class WrinklePatch {
	List pixels;
	List borderPixels;
	Vector3d color;
	
	public WrinklePatch(Wrinkle wrinkle, Image sourceImg){
		this.color = wrinkle.color;
		pixels = new ArrayList();
		
		int width = (int)(wrinkle.avgWidth * 1.3);
		int imageWidth = sourceImg.width;
		int imageHeight = sourceImg.height;
		boolean[][] visited = new boolean [imageWidth][imageHeight];
		for(int i = 0; i < visited.length; i++)
			for(int j = 0; j < visited.length; j++)
				visited[i][j] = false;
		
		for(int i = 0; i < wrinkle.pixels.size(); i++){
			//System.out.println("  looking at curve pixel: " + i);
			Pixel wrinkleCurvePixel = (Pixel)wrinkle.pixels.get(i);
			for(int dx = -width; dx <=width; dx++){
				int height = (int)Math.round(Math.sqrt(width*width - dx*dx));
				for(int dy = -height; dy <= height; dy++){
					int x = wrinkleCurvePixel.x + dx;
					int y = wrinkleCurvePixel.y + dy;
					if(x >= 1 && y >= 1 && 
					   x < imageWidth -1 && y < imageHeight - 1){
						Pixel pixel = new Pixel(x, y);
						if(/*!pixels.contains(pixel)*/ !visited[x][y]){
							pixels.add(pixel);
							visited[x][y] = true;
						}
					}
				}
			}
		}
		
		
		
		for(int i = 0; i < pixels.size(); i++){
			Pixel pixel = (Pixel)pixels.get(i);
			//pixel.height = wrinkle.diffImage.data[pixel.x][pixel.y];
		}

		getHeights(imageWidth, imageHeight, visited, width, sourceImg);
		
		

		//getHeights(imageWidth, imageHeight, visited, width);
		
	}
	
	public void getHeights(int imageWidth, int imageHeight, boolean[][] visited, int wrinkleWidth, Image sourceImg){
		boolean[][] border = new boolean[imageWidth][imageHeight];
		
		for(int i = 0; i < imageWidth; i++)
			for(int j = 0; j < imageHeight; j++)
				border[i][j] = false;

		for(int i = 0; i < imageWidth; i++){
			for(int j = 0; j < imageHeight; j++){
				if(!visited[i][j]){
					for(int dx = -1; dx <= 1; dx++){
						for(int dy = -1; dy <= 1; dy++){
							int x = i + dx;
							int y = j + dy;
							if(x < 0 || y < 0 || x >= imageWidth || y >= imageHeight)
								break;
							if(visited[x][y])
								border[i][j] = true;
						}
						
					}
				}
			}
		}
		
		for(int i = 0; i < imageWidth; i++){
			if(visited[i][0]) border[i][0] = true;
			if(visited[i][imageHeight - 1]) border[i][imageHeight - 1] = true;
		}
		for(int j = 0; j < imageHeight; j++){
			if(visited[0][j]) border[0][j] = true;
			if(visited[imageWidth - 1][j]) border[imageWidth - 1][j] = true;
		}
		
		int[][] distSq = new int[imageWidth][imageHeight];
		int[][] closestX = new int[imageWidth][imageHeight];
		int[][] closestY = new int[imageWidth][imageHeight];

		for(int i = 0; i < imageWidth; i++)
			for(int j = 0; j < imageHeight; j++)
				distSq[i][j] = Integer.MAX_VALUE;
		
		for(int i = 0 ; i < imageWidth; i++){
			for(int j = 0; j < imageHeight; j++){
				if(border[i][j]){
					for(int dx = -wrinkleWidth - 1; dx <= wrinkleWidth + 1; dx++)
						for(int dy = -wrinkleWidth - 1; dy <= wrinkleWidth + 1; dy++){
							int x = i + dx;
							int y = j + dy;
							if(x >= 0 && y >= 0 && x < imageWidth && y < imageHeight &&
									visited[x][y]){
								int currentDistSq = dx*dx + dy*dy;
								if(currentDistSq < distSq[x][y]){
									distSq[x][y] = currentDistSq;
									closestX[x][y] = i;
									closestY[x][y] = j;
								}
									
							}
								
						}
				}
			}
		}
		borderPixels = new ArrayList();
		for(int i = 0 ; i < imageWidth; i++)
			for(int j = 0; j < imageHeight; j++)
				if(border[i][j])
					borderPixels.add(new Pixel(i, j));

		float borderAvg = 0;
		for(int i = 0; i < borderPixels.size(); i++){
			Pixel pixel = (Pixel)borderPixels.get(i);
			borderAvg += sourceImg.data[pixel.x][pixel.y];
		}
		borderAvg /= borderPixels.size();
		
		float[][] smoothedInterior = solveLaplaceEq(pixels, sourceImg, borderAvg);
		
		for(int i = 0; i < pixels.size(); i++){
			Pixel pixel = (Pixel)pixels.get(i);
			int x = pixel.x;
			int y = pixel.y;
			int borderX = closestX[x][y];
			int borderY = closestY[x][y];
			float dist = (float)Math.sqrt((x - borderX) * (x - borderX) +
					 (y - borderY) * (y - borderY));
			if(dist > 40){
				System.out.println(dist + "!!!!!");
			}
			//pixel.height = dist;
			//pixel.height = sourceImg.data[borderX][borderY] - sourceImg.data[x][y];
			//pixel.height = borderAvg - sourceImg.data[x][y];
			pixel.height = smoothedInterior[pixel.x][pixel.y] - sourceImg.data[x][y];
		}													
	}
	
	static float[][] solveLaplaceEq(List interiorPixels, Image sourceImg, float borderAvg){
		float [][]result = new float[sourceImg.width][sourceImg.height];
		for(int i = 0; i < sourceImg.width; i++)
			for(int j = 0; j < sourceImg.height; j++)
				result[i][j] = sourceImg.data[i][j];

		for(int i = 0; i < interiorPixels.size(); i++){
			Pixel pixel = (Pixel)interiorPixels.get(i);
			int x = pixel.x;
			int y = pixel.y;
			result[x][y] = borderAvg;
		}		
		
		for(int iter = 0; iter < 80; iter++){
			float totalError = 0;
			for(int i = 0; i < interiorPixels.size(); i++){
				Pixel pixel = (Pixel)interiorPixels.get(i);
				int x = pixel.x;
				int y = pixel.y;
				float neighborAvg = .25f * (result[x + 1][y] + result[x - 1][y] + 
									result[x][y + 1] + result[x][y - 1]);
				totalError += Math.abs(neighborAvg - result[x][y]);
				result[x][y] = neighborAvg;
			}
			//System.out.println(iter + ": " + totalError);
		}
		return result;
	}
	
	public void draw(GL gl){
		gl.glBegin(GL.GL_POINTS);
		gl.glColor3d(color.x, color.y, color.z);
		for(Iterator iter = borderPixels.iterator(); iter.hasNext();){
			Pixel pixel = (Pixel)iter.next();
			gl.glVertex2d(pixel.x, pixel.y);
		}
		gl.glEnd();
	}
	
	static public List getPatches(List wrinkles, Image sourceImg){
		List patches = new ArrayList();
		for(int i = 0; i < wrinkles.size(); i++){
			System.out.println("  processing wrinkle: " + i);
			patches.add(new WrinklePatch((Wrinkle)wrinkles.get(i), sourceImg));
		}
		return patches;
	}

}
